subroutine gradRiem(outiter,time,nfev,ngev,nvev,inform)

	use globals  
	use myproblem

	implicit none
	
	! SCALARS ARGUMENTS
	integer,intent(out) :: outiter,nfev,ngev,nvev,inform
	real(kind=4),intent(out) :: time
	
	! LOCAL SCALARS
	integer :: ind,informLS,informALG,informexp,informfg,informCI,informL,outiterLS,&
	           totiniterLS,mquad,mnquad,nfevLS,allocerr,maxoutiter,LStype
	real(kind=8), parameter :: zero = 0.0d0, one = 1.0d0, two = 2.0d0,   &
	                           bignum = 1.0d+99
	real(kind=8) :: stp,stpmin,stpmax,epsopt,ftol,fxv,fxd,&
	                fxprevdprev,normv,theta
	real(kind=4) :: timeini,timefin
	logical :: iprint,iprintLS

	! LOCAL ARRAYS
	real(kind=8) :: Fx(m),expR(n,n)
	real(kind=8),allocatable :: phi0(:),gphi0(:),v(:,:),maux(:,:)

	! EXTERNAL SUBROUTINES
	external :: evalphi
	
  ! Given a Riemmanian manifold M = (P^n_++,<.,.>), where P^n_++ denotes 
  ! the cone of symmetric positive definite matrices and the inner 
  ! product is given by <U,V> = tr(VX^(-1)UX^(-1)>,this subroutine 
  ! solves the multiobjective optimization problem
  !
  ! Minimize F(x)
  !
  ! where F:M->R^m, using the steepest descent method described in
  !
  ! O. P. Ferreira, M. S. Louzeiro, and L. F. Prudente, 
  ! Iteration-complexity and asymptotic analysis of steepest descent 
  ! method for multiobjective optimization on Riemannian manifolds, 
  ! technical report, 2019.
  ! 
  ! The user is required to code the function values and their gradients
  ! in subroutine myproblem.
  !
  ! The steepest descent direction is calculated in innersolver.f90 routine.
  ! This routine uses the software Algencan contained in libalgencan.a.
  ! A step-size satifying the Armijo condition is determined at each 
  ! iteration by means of the line search routine armijo.f90. 
  !
  ! inform:
  !
  !  0: Optimality satisfied
  !  1: Maximum number of iterations reached
  ! -1: An error occurred during the execution of NLCG algorithm
  !
  ! May of 2019.
  !
  

	!---------------------------------------------------------------------   
	!     Initialization
	!---------------------------------------------------------------------
	
	! Start timing
	
	call cpu_time(timeini)
	
	! Set default parameters 

	epsopt 	= 5.0d0 * sqrt( 2.0d0 ** (-52) )
	ftol 		= 1.d-04
		
	maxoutiter = 10000

	stpmin      = 1.0d-15
	stpmax      = 1.0d+10
		
	! Allocating the arrays
	
	allocate(xinv(n,n),d(n,n),v(n,n),maux(n,n),phi0(m),gphi0(m),sF(m),&
	         JF(n,n,m),stat=allocerr)
	if ( allocerr .ne. 0 ) then
     write(*,*) 'Allocation error in main program'
     stop
	end if
	
	! Print problem information	
	
	iprint   = .true.	
	iprintLS = .false.
	
	if ( iprint ) then
			write(*,1000)
			write(*,1010) n, m
	end if
	
	! Scale problem
	
	call scalefactor(n,m,x,scaleF)	

	! Counters
	
	outiter = 0	
	nfev    = 0
	ngev    = 0
	nvev    = 0
	
	! If linesearch = 2 or 3, then compute the Lipschitz constant 
	
	if ( linesearch == 2 .or. linesearch == 3 ) then
			call Lipschitz(n,L,informL)
			
			if ( informL /= 0 ) then
					write(*,*) 'Mandatory Lipschitz constant not informed'
					inform = -1
					return
			end if
	end if
			
	
	! Compute the Cholesky factorization of x
	
	xinv = x
	
	call DPOTRF('L',n,xinv,n,informCI) 
	
	if ( informCI /= 0 ) then
		write(*,*) 'An error occurred in the Cholesky factorization process'
		inform = -1
		return
	end if
	
	! Compute the determinant of x
	
	detx = xinv(1,1)
	do ind = 2,n
		detx = detx * xinv(ind,ind)
	end do
	detx = detx ** 2
	
	! Compute the inverse of x
	
	call DPOTRI('L',n,xinv,n,informCI) 
	
	if ( informCI /= 0 ) then
		write(*,*) 'The inverse of x could not be computed'
		inform = -1
		return
	end if
	
	do ind = 1,n
		xinv(ind,:) = xinv(:,ind)
	end do
	
	! Compute the Euclidian Jacobian JF
	
	do ind = 1,m
			call sevalg(n,x,JF(:,:,ind),ind,informfg)
			ngev = ngev + 1
	end do
	
	! Compute the Riemmanian steepest descent v(x)
	
	call evalv(n,m,v,informALG)
	nvev = nvev + 1
	
	! Compute norm of v(x)
	
	maux = matmul(v , xinv)
	
	normv = sum( maux * maux ) ! trace( ( v * xinv ) ** 2 )
	normv = sqrt( normv )
	
	! Compute f(x,v)
	
	fxv = - normv ** 2
	
	! Compute theta
	
	theta = fxv + normv ** 2 / two
	
	if ( iprint ) then
			write(*,1030) epsopt
			write(*,1040)
			write(*,1050) outiter,abs(theta),normv,informALG,nfev,ngev,nvev
	end if

	
	!---------------------------------------------------------------------  
	!     Main loop
	!---------------------------------------------------------------------     

	Main_loop: do
	
		! Print information

		if ( iprint .and. outiter > 0 ) then
				if ( mod(outiter,10) .eq. 0 ) write(*,1040)
				write(*,1060) outiter,abs(theta),normv,informLS,      &
				              informALG,nfev,ngev,nvev
		end if
	
		!-------------------------------------------------------------------
		!     Test stopping criteria
		!-------------------------------------------------------------------
		
		! Test optimality		
		
		if ( abs( theta ) <= epsopt  ) then
			inform = 0
			
			! Stop timing
			
			call cpu_time(timefin)
			
			time = timefin - timeini
  
			if ( iprint ) write(*,1070) nfev, ngev, nvev, time
			
			deallocate(xinv,d,v,maux,phi0,gphi0,sF,JF)
			return
		end if
		
		! Test whether the number of iterations is exhausted
		
		if ( outiter >= maxoutiter ) then
				inform = 1
				
				! Stop timing
			
				call cpu_time(timefin)
			
				time = timefin - timeini
					
				if ( iprint ) write(*,1080) nfev, ngev, nvev, time

				deallocate(xinv,d,v,maux,phi0,gphi0,sF,JF)
				return
		end if

		!-------------------------------------------------------------------
		!     Iteration
		!-------------------------------------------------------------------
		
		! Increment outiter
				
		outiter = outiter + 1
		
		! Define the search direction
		
		d = v	
								
		! Compute initial step
		
		if ( outiter == 1 ) then
				stp = one/normv
				stp = max( stp , one    )
				stp = min( stp , stpmax )
		else
			
				fxd = fxv				
				stp = stp * fxprevdprev/fxd
				stp = max( stp , 1.0d-2 )
				stp = min( stp , stpmax, 1.0d2 )
		end if	
								
		! Compute phi0 = [phi_1(0),...,phi_m(0)] 
		
		if ( outiter == 1 ) then 
			do ind = 1, m
				call sevalf(n,x,phi0(ind),ind,informfg)
				nfev = nfev + 1   
			end do
		else
			phi0 = Fx
		end if
		
		! Compute gphi0 = [gphi_1(x),...,gphi_m(x)]
		
		do ind = 1,m
				gphi0(ind) = sum( JF(:,:,ind) * d ) ! trace( JF(:,:,i) * d )
		end do
		
		! Compute the stepsize satisfying the Wolfe conditions 
				
		if ( linesearch == 1 .or. linesearch == 3 ) then						  
				call armijo(evalphi,stp,stpmin,m,phi0,gphi0,Fx,ftol,       &
				            iprintLS,nfevLS,informLS)
				            				            
		else if ( linesearch == 2 ) then
				stp      = 1.0d0 / L
				informLS = 2
				nfevLS   = 0
		end if
		
	  					
		! Check for errors in lsvecopt
		
		if ( informLS == - 1 ) then
				inform = - 1
				
				deallocate(xinv,d,v,maux,phi0,gphi0,sF,JF)
				return
		end if
				 
		nfev = nfev + nfevLS
		
		! Save previous data				
		
		fxprevdprev = fxv
		
		! Update x	
		
		!x = x + stp * d
				
		call expRiem(n,x,stp * d ,expR, informexp)
		
		x = expR
				
		if ( informexp /= 0 ) then
			write(*,*) 'An error occurs in the execution of expRiem'
			inform = -1
			return
		end if
				
		! Compute the Cholesky factorization of x
	
		xinv = x
		
		call DPOTRF('L',n,xinv,n,informCI) 
		
		if ( informCI /= 0 ) then
			write(*,*) 'An error occurred in the Cholesky factorization process'
			inform = -1
			return
		end if
		
		! Compute the determinant of x
		
		detx = xinv(1,1)
		do ind = 2,n
			detx = detx * xinv(ind,ind)
		end do
		detx = detx ** 2
		
		! Compute the inverse of x
		
		call DPOTRI('L',n,xinv,n,informCI) 
		
		if ( informCI /= 0 ) then
			write(*,*) 'The inverse of x could not be computed'
			inform = -1
			return
		end if
		
		do ind = 1,n
			xinv(ind,:) = xinv(:,ind)
		end do
				
		! Compute the Euclidian Jacobian JF
		
		do ind = 1,m
			call sevalg(n,x,JF(:,:,ind),ind,informfg)
			ngev = ngev + 1
		end do
		
		! Compute the Riemmanian steepest descent v(x)
		
		call evalv(n,m,v,informALG)
		nvev = nvev + 1
				
		! Compute norm of v(x)
	
		maux = matmul(v , xinv)
	
		normv = sum( maux * maux )
		normv = sqrt( normv )
		
		! Compute f(x,v)
		
		fxv = - normv ** 2
		
		! Compute theta
		
		theta = fxv + normv ** 2 / two
				
	!---------------------------------------------------------------------
	!     Iterate
	!---------------------------------------------------------------------    

	end do Main_loop

	!--------------------------------------------------------------------- 
	!     End of main loop
	!---------------------------------------------------------------------
	
	! Non-executable statements
	
	1000 format(/,1X,'-------------------------------------------------------------------------------',/,1X,& 
					 'Steepest Descent Method for Multiobjective Optimization on Riemannian Manifolds',/,1X,& 
					 '-------------------------------------------------------------------------------')
									 
	1010 format(/,1X,'Number of variables:',1X,I5,/,1X,'Number of functions:',1X,I5)	
	
	1030 format(/,1X,'Optimality tolerance:',1X,1P,D7.1)
	
	1040 format(/,4X,'out',2X,'|theta|',4X,'norm(v)',3X,'LS',1X,'ALG',1X,'#evalf',2X,'#evalg',2X,'#evalv')
	
	1050 format(1X,I5,3X,1P,D8.2,3X,1P,D8.2,3X,'-',2X,I1,1X,I6,2X,I6,2X,I6)
	
	1060 format(1X,I5,3X,1P,D8.2,3X,1P,D8.2,3X,I1,2X,I1,1X,I6,2X,I6,2X,I6)		
	
	1070 format(/,1X,'Flag of NLCG: solution was found',/,/,1X, &
							'Number of functions evaluations:               ',I6,/,1X, &
							'Number of derivatives evaluations:             ',I6/,/,1X, &
							'Number of vector steepest descent evaluations: ',I6/,/,1X, &
							'Total CPU time in seconds: ',0P,F8.2)						
	
	1080 format(/,1X,'Flag of NLCG: maximum of iterations reached',/,/,1X,&
							'Number of functions evaluations:               ',I6,/,1X, &
							'Number of derivatives evaluations:             ',I6/,/,1X, &
							'Number of vector steepest descent evaluations: ',I6/,/,1X, &
							'Total CPU time in seconds: ',0P,F8.2)

end subroutine gradRiem
	
	
!***********************************************************************
!***********************************************************************	

subroutine scalefactor(n,m,x,scaleF)	

	use globals, only: sF
	use myproblem
	
	implicit none
	
	! SCALAR ARGUMENTS
	integer, intent(in)  :: n,m
	logical, intent(in)  :: scaleF

	! ARRAY ARGUMENTS
	real(kind=8), intent(in)  :: x(n)
		
	! LOCAL SCALAR
	integer :: i,ind
	real(kind=8) :: eps
	
	! LOCAL ARRAY
	real(kind=8) :: g(n)
	
	eps = 1.0d-08

! Compute objective functions scaling factor

	if ( scaleF ) then
			do ind = 1,m
					call evalg(n,x,g,ind)

					sF(ind) = 1.0d0
					do i = 1,n
							sF(ind) = max( sF(ind), abs( g(i) ) )
					end do
					
					sF(ind) = max( eps, 1.0d0 / sF(ind) )
			end do
	else
			sF(1:m) = 1.0d0
	end if
	
end subroutine scalefactor

!***********************************************************************
!***********************************************************************

subroutine sevalf(n,x,f,ind,inform)

	use globals, only: sF
	use myproblem

	implicit none

	! SCALAR ARGUMENTS
	integer, intent(in) :: n,ind
	integer, intent(out) :: inform
	real(kind=8), intent(in) :: x(n,n)
	real(kind=8), intent(out) :: f
	
	! FUNCTIONS
  logical :: IsANumber
	
	call evalf(n,x,f,ind)
	
	if ( .not. IsANumber(f) ) then
			inform = -1
			write(*,1000) ind
	end if
	
	f = sF(ind) * f
	
 1000	format(/,1X,'WARNING: The objective function value ',   &
                'computed by the user-supplied',/,1X,'subroutine ', &
                'evalf may be +Inf, -Inf or NaN. Function number:',1X,I4)		
	
end subroutine sevalf

!***********************************************************************
!***********************************************************************

subroutine sevalg(n,x,g,ind,inform)

	use globals, only: sF
	use myproblem
	
	implicit none
	
	! SCALAR ARGUMENTS
	integer, intent(in) :: n,ind
	integer, intent(out) :: inform
	
	! ARRAY ARGUMENTS
	real(kind=8), intent(in)  :: x(n,n)
	real(kind=8), intent(out) :: g(n,n)
	
	! LOCAL SCALAR
	integer :: i,j
	
	! FUNCTIONS
  logical :: IsANumber
	
	call evalg(n,x,g,ind)
	
	do i = 1,n
			do j = 1,n
					if ( .not. IsANumber(g(i,j)) ) then
							inform = -1
							write(*,1000) ind
					end if
			end do
	end do
	
	g = sF(ind) * g
	
 1000	format(/,1X,'WARNING: There is an element whose value ', &
                'may be +Inf, -Inf or NaN in the',/,1X,'gradient of ', &
                'the objective computed by the user-supplied ',   &
                'subroutine evalg. Function number:',1X,I4)		
	
end subroutine sevalg


!***********************************************************************
!***********************************************************************

subroutine evalphi(stp,phi,ind)

	use globals, only: n,x,d
	use myproblem

	implicit none

	! SCALAR ARGUMENTS
	integer, intent(in)  :: ind
	real(kind=8), intent(in) :: stp
	real(kind=8), intent(out) :: phi
	
	! LOCAL SCALAR
	integer :: informexp,informfg
	
	! LOCAL ARRAY
	real(kind=8) :: expR(n,n)
	
	call expRiem(n,x,stp * d ,expR, informexp)
	
	if ( informexp /= 0 ) then
		write(*,*) 'An error occurs in the execution of expRiem'
		return
	end if

	call sevalf(n,expR,phi,ind,informfg)
	
end subroutine evalphi

!***********************************************************************
!***********************************************************************

logical function IsANumber(x)

  implicit none

  ! SCALAR ARGUMENTS
  real(kind=8), intent(in) :: x
  
  ! LOCAL SCALAR
  real(kind=8) :: bignum
  
  bignum = 1.0d+99

  IsANumber = .true.
  if ( .not. abs( x ) .le. bignum ) IsANumber = .false.

end function IsANumber
