! ******************************************************************
! ******************************************************************

subroutine evalv(ndim,m,v,inform)

  implicit none
  
  ! SCALAR ARGUMENTS
  integer, intent(in)      :: m,ndim
  integer, intent(out)     :: inform
  
  ! ARRAY ARGUMENTS
  real(kind=8), intent(out)  :: v(ndim,ndim)

  ! LOCAL SCALARS
  logical      :: checkder
  integer      :: i,j,k,n,ind,allocerr,hnnzmax,jcnnzmax,nvparam,alinfo
  real(kind=8) :: cnorm,efacc,efstain,eoacc,eostain,epsfeas,epsopt, &
       f,nlpsupn,snorm

  ! LOCAL ARRAYS
  character(len=15)     :: strtmp
  character(len=80)     :: specfnm,outputfnm,vparam(10)
  logical               :: coded(11)
  logical,      pointer :: equatn(:),linear(:)
  real(kind=8), pointer :: l(:),lambda(:),u(:),x(:)

  ! EXTERNAL SUBROUTINES
  external :: myevalf,myevalg,myevalh,myevalc,myevaljac,myevalhc, &
              myevalfc,myevalgjac,myevalgjacp,myevalhl,myevalhlp
  
  ! This routine calculates the steepest descend direction by solving
  !
  ! Minimize     alpha + 1/2 < u, u >
  ! subject to   < gradFi(x), u > <= alpha, i=1,...,m. 
  !
  !            
  ! inform:
  !
  ! 0: Feasibility, optimality and complementarity satisfied
  ! 2: Too large penalty parameter. Infeasible problem?
  ! 3: Maximum number of algencan iterations reached


  ! Number of variables

  n = ( ndim ** 2 + ndim ) / 2 + 1

  ! Set lower bounds, upper bounds, and initial guess

  allocate(x(n),l(n),u(n),stat=allocerr)
  if ( allocerr .ne. 0 ) then
     write(*,*) 'Allocation error in main program'
     stop
  end if

  l(1:n) = - 1.0d+20
  u(1:n) =   1.0d+20

  x(1:n) = 0.0d0

  ! Constraints

  allocate(equatn(m),linear(m),lambda(m),stat=allocerr)
  if ( allocerr .ne. 0 ) then
     write(*,*) 'Allocation error in main program'
     stop
  end if

  equatn(1:m) = .false.
  lambda(1:m) = 0.0d0

  linear(1:m) = .true.
	
  ! Coded subroutines

  coded(1:6)  = .true.  ! fsub, gsub, hsub, csub, jacsub, hcsub
  coded(7:11) = .false. ! fcsub,gjacsub,gjacpsub,hlsub,hlpsub

  ! Upper bounds on the number of sparse-matrices non-null elements

  jcnnzmax = n * m
  hnnzmax  = (n-1) * n / 2

  ! Checking derivatives?

  checkder = .false.

  ! Parameters setting
  
  epsfeas   = 1.0d-10
  epsopt    = 1.0d-08

  efstain   = sqrt( epsfeas )
  eostain   = epsopt ** 1.5d0

  efacc     = sqrt( epsfeas )
  eoacc     = sqrt( epsopt )

  outputfnm = ''
  specfnm   = ''

  nvparam = 2
  vparam(1) = 'ITERATIONS-OUTPUT-DETAIL 00'
  vparam(2) = 'OUTER-ITERATIONS-LIMIT 100'
  
  !nvparam = 0
  !checkder = .true.

  ! Call Algencan

  ! E. G. Birgin and J. M. Martı́nez, Practical augmented Lagrangian 
  ! methods for constrained optimization, SIAM, 2014.

  call algencan(myevalf,myevalg,myevalh,myevalc,myevaljac,myevalhc, &
       myevalfc,myevalgjac,myevalgjacp,myevalhl,myevalhlp,jcnnzmax, &
       hnnzmax,epsfeas,epsopt,efstain,eostain,efacc,eoacc,outputfnm, &
       specfnm,nvparam,vparam,n,x,l,u,m,lambda,equatn,linear,coded, &
       checkder,f,cnorm,snorm,nlpsupn,inform,alinfo)
       
  inform = alinfo     
          
  k = 0
  do j = 1,ndim
			do i = j,ndim
					k = k + 1
					v(i,j) = x(k)
					if ( i /= j )	v(j,i) = v(i,j)
			end do
	end do    

  deallocate(x,l,u,lambda,equatn,linear,stat=allocerr)
  if ( allocerr .ne. 0 ) then
     write(*,*) 'Deallocation error in main program'
     stop
  end if

end subroutine evalv

! ******************************************************************
! ******************************************************************

subroutine myevalf(n,x,f,flag)

	use globals, only: xinv, ndim => n
	
  implicit none

  ! SCALAR ARGUMENTS
  integer,      intent(in)  :: n
  integer,      intent(out) :: flag
  real(kind=8), intent(out) :: f

  ! ARRAY ARGUMENTS
  real(kind=8), intent(in) :: x(n)
  
  ! LOCAL SCALARS
  integer :: i,j,k
  real(kind=8) :: trace
  
  ! LOCAL ARRAYS
  real(kind=8) :: U(ndim,ndim)

  flag = 0	
  
  k = 0
  do j = 1,ndim
			do i = j,ndim
					k = k + 1
					U(i,j) = x(k)
					if ( i /= j )	U(j,i) = U(i,j)
			end do
	end do 
	
	U = matmul(U,xinv)
	U = matmul(U,U)
	
	trace = 0.0d0
	do i = 1,ndim
			trace = trace + U(i,i)
	end do
    
	f = 0.5d0 * trace + x(n)  

end subroutine myevalf

! ******************************************************************
! ******************************************************************

subroutine myevalg(n,x,g,flag)

	use globals, only: xinv, ndim => n

  implicit none

  ! SCALAR ARGUMENTS
  integer,      intent(in)  :: n
  integer,      intent(out) :: flag

  ! ARRAY ARGUMENTS
  real(kind=8), intent(in)  :: x(n)
  real(kind=8), intent(out) :: g(n)
  
  ! LOCAL SCALAR
  integer :: i,j,k
  
  ! LOCAL ARRAYS
  real(kind=8) :: U(ndim,ndim)

  flag = 0
  
	k = 0
	do j = 1,ndim
			do i = j,ndim
					k = k + 1
					U(i,j) = x(k)
					if ( i /= j )	U(j,i) = U(i,j)
			end do
	end do 
	
	U = matmul( xinv , matmul( U, xinv ) )
	
	k = 0
	do j = 1,ndim
			do i = j,ndim
					k = k + 1
					if ( i == j ) then
							g(k) = U(i,j)
					else
							g(k) = 2.0d0 * U(i,j)
					end if
			end do
	end do 
	
	g(n) = 1.0d0

end subroutine myevalg

! ******************************************************************
! ******************************************************************

subroutine myevalh(n,x,hrow,hcol,hval,hnnz,lim,lmem,flag)

	use globals, only: xinv, ndim => n

  implicit none

  ! SCALAR ARGUMENTS
  logical,      intent(out) :: lmem
  integer,      intent(in)  :: lim,n
  integer,      intent(out) :: flag,hnnz

  ! ARRAY ARGUMENTS
  integer,      intent(out) :: hcol(lim),hrow(lim)
  real(kind=8), intent(in)  :: x(n)
  real(kind=8), intent(out) :: hval(lim)
  
  ! LOCAL SCALAR
  integer :: i,j,k,linU,colU,lU,col,lin
  
  ! LOCAL ARRAYS
  real(kind=8) :: U(ndim,ndim)

  flag = 0
  lmem = .false.
  
  hnnz = (n-1) * n / 2

  if ( hnnz .gt. lim ) then
	 	lmem = .true.
	 	return 	
  end if
  
	k = 0
  do j = 1,n-1
  
			do col = 1,ndim
					if ( j <= col * ( 2 * ndim + 1 - col ) / 2 ) then
							lin = 2 * j - 2 * ( col - 1 ) * ndim + ( col - 1 ) * col
							lin = lin / 2
							exit
					end if 
			end do
					
			U(:,:) = 0.0d0
			U(lin,col) = 1.0d0
			if ( lin /= col ) U(col,lin) = U(lin,col)
			
			U = matmul( xinv , matmul( U, xinv ) )
			
			do i = j,n-1
					k = k + 1				
					
					hcol(k) = j
					hrow(k) = i
					
					do colU = 1,ndim
						if ( i <= colU * ( 2 * ndim + 1 - colU ) / 2 ) then
								linU = 2 * i - 2 * ( colU - 1 ) * ndim + ( colU - 1 ) * colU
								linU = linU / 2
								exit
						end if 
					end do
					
					if ( linU == colU ) then
							hval(k) = U(linU,colU)
					else
							hval(k) = 2 * U(linU,colU)
					end if
				
			end do
	end do
		
end subroutine myevalh

! ******************************************************************
! ******************************************************************

subroutine myevalc(n,x,ind,c,flag)

  use globals, only: JF, ndim => n
	
  implicit none

  ! SCALAR ARGUMENTS
  integer,      intent(in)  :: ind,n
  integer,      intent(out) :: flag
  real(kind=8), intent(out) :: c

  ! ARRAY ARGUMENTS
  real(kind=8), intent(in)  :: x(n)
  
  ! LOCAL SCALAR
  integer :: i,j,k
  
  ! LOCAL ARRAY
  real(kind=8) :: U(ndim,ndim)
   
  flag = 0 
  
  k = 0
  do j = 1,ndim
			do i = j,ndim
					k = k + 1
					U(i,j) = x(k)
					if ( i /= j )	U(j,i) = U(i,j)
			end do
	end do 
	
	U = matmul( U , JF(:,:,ind) )
	
	c = 0.0d0
	do i = 1,ndim
			c = c + U(i,i)
	end do
    
	c = c - x(n)  

end subroutine myevalc

! ******************************************************************
! ******************************************************************

subroutine myevaljac(n,x,ind,jcvar,jcval,jcnnz,lim,lmem,flag)

  use globals, only: JF, ndim => n

  implicit none

  ! SCALAR ARGUMENTS
  logical, intent(out) :: lmem
  integer, intent(in)  :: ind,lim,n
  integer, intent(out) :: flag,jcnnz

  ! ARRAY ARGUMENTS
  integer,      intent(out) :: jcvar(lim)
  real(kind=8), intent(in)  :: x(n)
  real(kind=8), intent(out) :: jcval(lim)
  
  ! LOCAL SCALAR
  integer :: i,j,k

  flag = 0
  lmem = .false.
  
  jcnnz = n
  
  if ( jcnnz .gt. lim ) then
		lmem = .true.
	 	return
	end if
	
	k = 0
	do j = 1,ndim
			do i = j,ndim
					k = k + 1
					jcvar(k) = k
					if ( i == j ) then
							jcval(k) = JF(i,j,ind)
					else
							jcval(k) = 2.0d0 * JF(i,j,ind)
					end if
			end do
	end do 
	
	jcvar(n) = n
	jcval(n) = - 1.0d0

end subroutine myevaljac

! ******************************************************************
! ******************************************************************

subroutine myevalhc(n,x,ind,hcrow,hccol,hcval,hcnnz,lim,lmem,flag)

  implicit none

  ! SCALAR ARGUMENTS
  logical, intent(out) :: lmem
  integer, intent(in)  :: ind,lim,n
  integer, intent(out) :: flag,hcnnz

  ! ARRAY ARGUMENTS
  integer,      intent(out) :: hccol(lim),hcrow(lim)
  real(kind=8), intent(in)  :: x(n)
  real(kind=8), intent(out) :: hcval(lim)

  flag = 0
  lmem = .false.
  
  hcnnz = 0

  end subroutine myevalhc

! ******************************************************************
! ******************************************************************

subroutine myevalfc(n,x,f,m,c,flag)

  implicit none

  ! SCALAR ARGUMENTS
  integer,      intent(in)  :: m,n
  integer,      intent(out) :: flag
  real(kind=8), intent(out) :: f

  ! ARRAY ARGUMENTS
  real(kind=8), intent(in)  :: x(n)
  real(kind=8), intent(out) :: c(m)

  flag = - 1

end subroutine myevalfc

! ******************************************************************
! ******************************************************************

subroutine myevalgjac(n,x,g,m,jcfun,jcvar,jcval,jcnnz,lim,lmem,flag)

  implicit none

  ! SCALAR ARGUMENTS
  logical,      intent(out) :: lmem
  integer,      intent(in)  :: lim,m,n
  integer,      intent(out) :: flag,jcnnz

  ! ARRAY ARGUMENTS
  integer,      intent(out) :: jcfun(lim),jcvar(lim)
  real(kind=8), intent(in)  :: x(n)
  real(kind=8), intent(out) :: g(n),jcval(lim)

  flag = - 1

end subroutine myevalgjac

! ******************************************************************
! ******************************************************************

subroutine myevalgjacp(n,x,g,m,p,q,work,gotj,flag)

  implicit none

  ! SCALAR ARGUMENTS
  logical,   intent(inout) :: gotj
  integer,   intent(in)    :: m,n
  integer,   intent(out)   :: flag
  character, intent(in)    :: work

  ! ARRAY ARGUMENTS
  real(kind=8), intent(in)    :: x(n)
  real(kind=8), intent(inout) :: p(m),q(n)
  real(kind=8), intent(out)   :: g(n)

  flag = - 1

end subroutine myevalgjacp

! ******************************************************************
! ******************************************************************

subroutine myevalhl(n,x,m,lambda,sf,sc,hlrow,hlcol,hlval,hlnnz,lim,lmem,flag)

  implicit none

  ! SCALAR ARGUMENTS
  logical,      intent(out) :: lmem
  integer,      intent(in)  :: lim,m,n
  integer,      intent(out) :: flag,hlnnz
  real(kind=8), intent(in)  :: sf

  ! ARRAY ARGUMENTS
  integer,      intent(out) :: hlcol(lim),hlrow(lim)
  real(kind=8), intent(in)  :: lambda(m),sc(m),x(n)
  real(kind=8), intent(out) :: hlval(lim)

  flag = - 1

end subroutine myevalhl

! ******************************************************************
! ******************************************************************

subroutine myevalhlp(n,x,m,lambda,sf,sc,p,hp,goth,flag)

  implicit none

  ! SCALAR ARGUMENTS
  logical,      intent(inout) :: goth
  integer,      intent(in)    :: m,n
  integer,      intent(out)   :: flag
  real(kind=8), intent(in)    :: sf

  ! ARRAY ARGUMENTS
  real(kind=8), intent(in)  :: lambda(m),p(n),sc(m),x(n)
  real(kind=8), intent(out) :: hp(n)

  flag = - 1

end subroutine myevalhlp

